/*
 * Copyright 1999, 2001 Silicon Graphics, Inc.
 * Copyright 2001 Ralf Baechle
 *           2001-04 Guido Guenther <agx@sigxcpu.org>
 */
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#include <prom.h>
#include <elf.h>

#include <sys/types.h>
#include <ext2_fs.h>
#include <ext2fs.h>

#include "arcboot.h"

#include <subarch.h>

#define ANSI_CLEAR	"\033[2J"
#define CONF_FILE	"/etc/arcboot.conf"

static int is64=0;

typedef	union {
	unsigned char e_ident[EI_NIDENT];
	Elf32_Ehdr header32;
	Elf64_Ehdr header64;
} Elf_Ehdr;

int LoadProgramSegments32(ext2_file_t file, Elf_Ehdr * header, void *segments)
{
	int idx;
	int loaded = 0;
	Elf32_Phdr* segment=(Elf32_Phdr*)segments;
	errcode_t status;
	size_t size;

	printf("Loading 32-bit executable\n\r");

	for (idx = 0; idx < header->header32.e_phnum; idx++) {

		if (segment->p_type == PT_LOAD && segment->p_vaddr) {
		    printf
			("Loading program segment %u at 0x%x, offset=0x%x, size = 0x%x\n\r",
		        idx + 1, KSEG0ADDR(segment->p_vaddr),
		        segment->p_offset, segment->p_filesz);

		    status =
			ext2fs_file_lseek(file, segment->p_offset,
				        EXT2_SEEK_SET, NULL);
		    if (status != 0) {
			    print_ext2fs_error(status);
			    prom_fatal("Cannot seek to program segment\n\r");
		    }

		    arc_do_progress = 1;
		    status = ext2fs_file_read(file,
					  (void *) (KSEG0ADDR(
					            segment->p_vaddr)),
					  segment->p_filesz, NULL);
		    printf("\n\n\r");	/* Clear progress */
		    arc_do_progress = 0;
		    if (status != 0) {
			print_ext2fs_error(status);
			prom_fatal("Cannot read program segment\n\r");
		    }

		    size = segment->p_memsz - segment->p_filesz;
		    if (size > 0) {
			    printf
				("Zeroing memory at 0x%x, size = 0x%x\n\r",
			        (KSEG0ADDR(segment->p_vaddr +
			        segment->p_filesz)), size);
			    memset((void *)
			       (KSEG0ADDR(segment->
				 p_vaddr + segment->p_filesz)), 0, size);
		    }

		    loaded = 1;
	    }

	    segment =
		(Elf32_Phdr *) (((char *) segment) +
			    header->header32.e_phentsize);
    }

	return loaded;
}

int LoadProgramSegments64(ext2_file_t file, Elf_Ehdr * header, void *segments)
{
	int idx;
	int loaded = 0;
 	Elf64_Phdr* segment=(Elf64_Phdr*)segments;
	errcode_t status;
	size_t size;

	is64=1;
	printf("Loading 64-bit executable\n\r");

	for (idx = 0; idx < header->header64.e_phnum; idx++) {
		if (segment->p_type == PT_LOAD) {
		    printf ("Loading program segment %u at 0x%x, "
			    "offset=0x%lx %lx, size = 0x%lx %lx\n\r",
			    idx + 1,
			    (int)KSEG0ADDR(segment->p_vaddr),
			    (long)(segment->p_offset>>32),
			    (long)(segment->p_offset&0xffffffff),
			    (long)(segment->p_filesz>>32),
			    (long)(segment->p_filesz&0xffffffff));

		    status =
			ext2fs_file_lseek(file, segment->p_offset,
				        EXT2_SEEK_SET, NULL);
		    if (status != 0) {
			    print_ext2fs_error(status);
			    prom_fatal("Cannot seek to program segment\n\r");
		    }

		    arc_do_progress = 1;
		    status = ext2fs_file_read(file,
					  (void *) (KSEG0ADDR(
					            segment->p_vaddr)),
					  segment->p_filesz, NULL);
		    arc_do_progress = 0;
		    if (status != 0) {
			print_ext2fs_error(status);
			prom_fatal("Cannot read program segment\n\r");
		    }

		    size = segment->p_memsz - segment->p_filesz;
		    if (size > 0) {
			    printf
				("Zeroing memory at 0x%lx, size = 0x%lx\n\r",
			        (KSEG0ADDR(segment->p_vaddr +
			        segment->p_filesz)), size);
			    memset((void *)
			       (KSEG0ADDR(segment->
				 p_vaddr + segment->p_filesz)), 0, size);
		    }

		    loaded = 1;
		}

		segment = (Elf64_Phdr *) (((char *) segment)
					  + header->header64.e_phentsize);
	}

	return loaded;
}

void LoadProgramSegments(ext2_file_t file, Elf_Ehdr * header)
{
	Boolean loaded = False;
	void *segments;
	size_t size;
	errcode_t status;

	if (header->e_ident[EI_CLASS] == ELFCLASS32)
		size = (size_t) (header->header32.e_phentsize *
				 header->header32.e_phnum);
	else
		size = (size_t) (header->header64.e_phentsize *
				 header->header64.e_phnum);

	if (size <= 0)
		prom_fatal("No program segments\n\r");

	segments = malloc(size);
	if (segments == NULL)
		prom_fatal("Cannot allocate memory for segment headers\n\r");
	else
		printf("Allocated 0x%x bytes for segments\n\r",size);

	if (header->e_ident[EI_CLASS] == ELFCLASS32) {
		status = ext2fs_file_lseek(file,
				(ext2_off_t) header->header32.e_phoff,
				EXT2_SEEK_SET, NULL);
	} else {
		status = ext2fs_file_lseek(file,
				(ext2_off_t) header->header64.e_phoff,
				EXT2_SEEK_SET, NULL);
	}
	if (status != 0) {
		print_ext2fs_error(status);
		prom_fatal("Cannot seek to program segment headers\n\r");
	}

	status = ext2fs_file_read(file, segments, size, NULL);
	if (status != 0) {
		print_ext2fs_error(status);
		prom_fatal("Cannot read program segment headers\n\r");
	}

	if(header->e_ident[EI_CLASS] == ELFCLASS32)
		loaded = LoadProgramSegments32(file, header, segments);
	else
		loaded = LoadProgramSegments64(file, header, segments);

	if (!loaded)
		prom_fatal("No loadable program segments found\n\r");

	free(segments);
}


Elf64_Addr LoadKernelFile(ext2_file_t file)
{
	Elf_Ehdr header;
	Elf64_Addr entry;
	errcode_t status;

	status =
	    ext2fs_file_read(file, (void *) &header, sizeof(header), NULL);
	if (status != 0) {
		print_ext2fs_error(status);
		prom_fatal("Can't read file header\n\r");
	}

	if (memcmp(&(header.e_ident[EI_MAG0]), ELFMAG, SELFMAG) != 0)
		prom_fatal("Not an ELF file\n\r");

	if (header.e_ident[EI_CLASS] != ELFCLASS32 &&
	    header.e_ident[EI_CLASS] != ELFCLASS64)
		prom_fatal("Not a 32-bit or 64-bit file\n\r");

	if (header.e_ident[EI_DATA] != ELFDATA2MSB)
		prom_fatal("Not a big-endian file\n\r");

	if (header.e_ident[EI_VERSION] != EV_CURRENT)
		prom_fatal("Wrong ELF version\n\r");

	if (header.e_ident[EI_CLASS]==ELFCLASS32) {
	    if (header.header32.e_type != ET_EXEC)
		prom_fatal("Not an executable file\n\r");

	    if (header.header32.e_machine != EM_MIPS)
		prom_fatal("Unsupported machine type\n\r");

	    if (header.header32.e_version != EV_CURRENT)
		prom_fatal("Wrong ELF version\n\r");

 	    entry = (Elf64_Addr) header.header32.e_entry;
	} else {
	    if (header.header64.e_type != ET_EXEC)
		prom_fatal("Not an executable file\n\r");

	    if (header.header64.e_machine != EM_MIPS)
		prom_fatal("Unsupported machine type\n\r");

	    if (header.header64.e_version != EV_CURRENT)
		prom_fatal("Wrong ELF version\n\r");

 	    entry = header.header64.e_entry;
	}

	LoadProgramSegments(file, &header);

	return entry;
}


Boolean OpenFile(const char *partition, const char *filename, ext2_file_t* file)
{
	extern io_manager arc_io_manager;
	ext2_filsys fs;
	ext2_ino_t file_inode;
	errcode_t status;

	initialize_ext2_error_table();

	status = ext2fs_open(partition, 0, 0, 0, arc_io_manager, &fs);
	if (status != 0) {
		print_ext2fs_error(status);
		return False;
	}

	status = ext2fs_namei_follow
	    (fs, EXT2_ROOT_INO, EXT2_ROOT_INO, filename, &file_inode);
	if (status != 0) {
		print_ext2fs_error(status);
		return False;
	}

	status = ext2fs_file_open(fs, file_inode, 0, file);
	if (status != 0) {
		print_ext2fs_error(status);
		return False;
	}
	return True;
}

Elf64_Addr LoadKernel(const char *partition, const char *filename)
{
	ext2_file_t file;

	if(!OpenFile( partition, filename, &file ))
		prom_fatal("Can't load kernel!\n\r");
	return LoadKernelFile(file);
}

void printCmdLine(int argc, char *argv[])
{
 int i;

 for(i = 0; i < argc; i++ )
	printf("%u: %s\n\r", i, argv[i]);
}

void _start64(long argc, char * argv[], char * envp[],
	      unsigned long long *addr)
{
	__asm__ __volatile__(
		".set push\n"
		"\t.set mips3\n"
		"\t.set noreorder\n"
		"\t.set noat\n"
		"\tld $1, 0($7)\n"
		"\tjr $1\n"
		"\t nop\n"
		"\t.set pop");
}

void _start(long argc, char *argv[], char *envp[])
{
	char	**kargv,
		**params,
		**targv;
	char	*ospartition,
		*oslabel,
		*osoptions,
		*kernelfile;
	long	kargc, targc, i;

	Elf32_Addr kernel_entry32;
	Elf64_Addr kernel_entry64;

	prom_init();

	/* Print identification */
	printf(ANSI_CLEAR "\n\rarcsboot: ARCS Linux ext2fs loader "
					__ARCSBOOT_VERSION__ "\n\n\r");

	prom_init_malloc(KERNELADDR, MAXLOADSIZE);
	prom_parse_arg(argc, argv, envp);

	ospartition=prom_get_ospartition();
	oslabel=prom_get_oslabel();

	if (!ospartition)
		prom_fatal("No ospartition\n");
	if (!oslabel)
		oslabel="Linux";

	params=ReadConfFile(&ospartition, CONF_FILE, oslabel);

	if (!params) {
		printf("Couldn't find label: %s in %s.\n\r", oslabel, CONF_FILE);
		printf("Will try to boot %s%s.\n\r", ospartition, oslabel);

		/* Get command line */
		prom_get_arg(&kargc, &kargv);
		kernelfile=oslabel;
	} else {
		kernelfile=params[1];

		/* Count kernel command line */
		kargv=&params[1];
		for(kargc=0;kargv[kargc];kargc++);

		/* Get possible environment vars with kernel options and append */
		osoptions=prom_get_osoptions();
		if (osoptions)
			kargv[kargc++]=osoptions;

		/* Get args and append to kernel command line */
		prom_get_arg(&targc, &targv);
		for(i=0;i<targc;i++)
			kargv[kargc++]=targv[i];
	}

	printf("Loading %s from %s\n\r", (params) ? params[0] : kernelfile, ospartition);

	kernel_entry64 = LoadKernel(ospartition, kernelfile);
	kernel_entry32 = (Elf32_Addr) kernel_entry64;

	if( kernel_entry64 ) {
	    if(is64==0){
		printf("Starting ELF32 kernel\n\r");
		prom_flush_all_caches();
		((void (*)(int argc, char * argv[], char * envp[]))
		    kernel_entry32)(kargc, kargv, envp);
	    } else {
		printf("Starting ELF64 kernel\n\r");
		prom_flush_all_caches();
		_start64(kargc, kargv, envp, &kernel_entry64);
	    }
	} else {
		printf("Invalid kernel entry NULL\n\r");
	}

	/* Not likely to get back here in a functional state, but what the heck */
	prom_wait("\n\r--- Press <spacebar> to restart ---");
	prom_restart();
}
