
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "sniprom.h"

static int	iargc;
static char	**iargv;

#define uint8_t	char

void dump_hex(uint8_t *buf, int size) {
	int		i;
	unsigned char	ch;
	char		sascii[17];

	sascii[16]=0x0;

	for(i=0;i<size;i++) {
		ch=buf[i];
		if (i%16 == 0) {
			printf("%x ", i);
		}
		printf("%x ", ch);
		if (ch >= ' ' && ch <= '}')
			sascii[i%16]=ch;
		else
			sascii[i%16]='.';

		if (i%16 == 15)
			printf("   %s\n\r", sascii);
	}

	/* i++ after loop */
	if (i%16 != 0) {

		for(;i%16 != 0;i++) {
			printf("   ");
			sascii[i%16]=' ';
		}

		printf("   %s\n\r", sascii);
	}
	printf("\n\r");
}


void prom_wait(const char *prompt) {
	printf("%s: called\n\r", __FUNCTION__);
}

void prom_fatal(const char *message, ...) {
	va_list ap;

	if (message != NULL) {
		printf("FATAL ERROR:  ");
		va_start(ap, message);
		vprintf(message, ap);
		va_end(ap);
	}

	prom_wait("\n\r--- Press <spacebar> to enter ARC interactive mode ---");
	__prom_reset();
}

int (*__prom_lseek)(FILE, int, int)   = NULL;


/*
 * The SNI Prom does not contain an official entry point for "prom_seek" which is
 * a pain and basically is a showstopper for bootloaders. There is a seek function
 * though which seems to be inbetween prom_open and prom_read. We try to find
 * it by looking for the function prolog e.g. the $sp setup. We then know the
 * address of the seek.
 */

void prom_init(void ) {
	unsigned int	jump;
	unsigned int	*readaddr,
			*openaddr,
			*saddr;

	jump = *((unsigned int *)PROM_ENTRY(PROM_READ));
	readaddr=(unsigned int *) (((jump & 0x03ffffff)<<2)|0xb0000000);
	jump = *((unsigned int *)PROM_ENTRY(PROM_OPEN));
	openaddr= (unsigned int *) (((jump & 0x03ffffff)<<2)|0xb0000000);

	for(saddr=readaddr-1;saddr>openaddr;saddr--) {
		/* Search for addiu $sp, negative */
		if ((*saddr & 0xffff8000) == 0x27bd8000) {
			__prom_lseek=(void *) saddr;
			break;
		}
	}

	if (__prom_lseek == NULL)
		prom_fatal("Didnt find prom_seek prolog between open and read\n\r");
}

void prom_parse_arg(int argc, char **argv, char **envp) {
	int	i;

	iargc=argc;
	iargv=argv;
}

char *prom_get_ospartition(void ) {
	return "dkncr(0,0,1)";
}

char *prom_get_oslabel(void ) {
	return "linux";
}

/* SNI Prom doesnt have OSOptions */
char *prom_get_osoptions(void ) {
	return NULL;
}

void prom_get_arg(long *argc, char ***argv) {
	*argc=iargc-1;
	*argv=iargv;
}

void prom_init_malloc(unsigned long loadaddr, unsigned long size) {
	MEMCONF		mc;
	unsigned long	stack = (unsigned long) &mc;
	int		i;

	__prom_get_memconf(&mc);

	for(i=0;i<8 && mc.bank[i].size;i++) {
		unsigned long start = KSEG0ADDR(mc.bank[i].base);
		unsigned long end = start + mc.bank[i].size;

		/* Leave some space for our stack */
		if ((stack >= start) && (stack < end))
			end =
			    (stack -
			     (STACK_PAGES *
			      PAGE_SIZE)) & ~(PAGE_SIZE - 1);

		/* Don't use memory from reserved region */
		if ((start >= loadaddr ) && (start < (loadaddr + size)))
			start = loadaddr + size;
		if ((end > loadaddr) && (end <= (loadaddr + size))) 
			end = loadaddr;
		if (end > start) {
			start+=0x80000;
			malloc_area_add(start, end - start);
		}

	}
}

int prom_seek(FILE stream, long long position, int whence) {
	int	pos32=position, res;

	res=__prom_lseek(stream, pos32, whence);

	if (res != position)
		prom_fatal("Seek failed\n\r");

	return 0;
}

void prom_restart(void ) {
	printf("%s: called\n\r", __FUNCTION__);
}

void prom_flush_all_caches(void ) {
	__prom_flushcache();
}

void prom_return_interactive(void ) {
	printf("%s: called\n\r", __FUNCTION__);
}

int prom_write(FILE stream, char *buf, unsigned long len, unsigned long *rlen) {
	int	res;
	/* SNI returns no status but standard conform length */
	res=__prom_write(stream, buf, len);
	*rlen=res;
	return res<0;
}

/*
 * We have seen larger reads to fail on RM200C Prom Version 5
 * and RM400 Prom Version 4. It seems this is secondary cache related. As we
 * would like to use the read-ahead and cache we simply split the reads into
 * smaller chunks here.
 *
 */
int prom_read(FILE stream, char *buf, unsigned long len, unsigned long *rlen) {
	int	res=0,i;
	int	read=0, thisread;

	while(read < len) {
		int left=len-read;
		if (left > 4096)
			thisread=4096;
		else
			thisread=left;

		res=__prom_read(stream, buf+read, thisread);

		read+=thisread;

		if (res < thisread)
			break;
	}
	*rlen=read;

	return (len > read);
}

int prom_open(char *name, int mode, FILE *stream) {
	int	res;
	res=__prom_open(name, mode);
	*stream=res;
#if 0
	{
		int i;
		uint8_t	sec[512];
		unsigned long rlen;
		for(i=0;i<64;i++) {
			prom_seek(*stream, i*512, 0);
			prom_read(*stream, (void *) &sec, 512, &rlen);
			printf("Sector %u\n\r", i);
			dump_hex(sec, 64);
		}

	}
#endif

	return 0;
}

int prom_close(FILE stream) {
	return 0;
}

